
# Imports ----------------------------------------------------------------------
import os.path
from shutil import which
from typing import List, Union


# Base functions ---------------------------------------------------------------
def is_convertable_to(input_object, input_type) -> bool:
    """Return if input_object is convertable to input_type."""
    try:
        _ = input_type(input_object)
        return True
    except:
        return False
    
def memory_str(n_bytes: int) -> str:
    """Return a human readable string for a memory size measure (input in bytes)."""
    if n_bytes / 1000**3 > 1.0:
        return f"{n_bytes / 1000**3:.3f} GB"
    elif n_bytes / 1000**2 > 1.0:
        return f"{n_bytes / 1000**2:.3f} MB"
    elif n_bytes / 1000 > 1.0:
        return f"{n_bytes / 1000:.3f} kB"
    else:
        return f"{n_bytes} B"
    
def time_str(n_sec: float) -> str:
    """Return a human readable string for a time measure (input in seconds)."""
    if n_sec / (60*60*24) > 1.0:
        return f"{n_sec / (60*60*24):.3f} d."
    elif n_sec / (60*60) > 1.0:
        return f"{n_sec / (60*60):.3f} h."
    elif n_sec / 60 > 1.0:
        return f"{n_sec / 60:.3f} min."
    else:
        return f"{n_sec:.3f} sec."
    
def find_file(path_list: List[str], is_software: bool, name: str, description: Union[str, None]=None, verbose: bool=False,) -> str:
    """Find first existing file among path_list."""
    
    # Find valid path among candidates
    output_path = None
    for candidate_path in path_list:

        # Find as a path to a file
        if os.path.isfile(candidate_path):
            output_path = candidate_path
            if verbose:
                print(f" * Set path for [{name}] (AS PATH TO A FILE): '{output_path}'")
            break

    # Find valid bash command executable in PATH
    if output_path is None and is_software:
        for candidate_path in path_list:
            basename = os.path.basename(candidate_path)
            which_candidate_path = which(basename)
            if which_candidate_path is not None:
                output_path = which_candidate_path
                if verbose:
                    print(f" * set path for [{name}] (AS EXECUTABLE): '{output_path}'")
                break

    # Raise error if no valid path is found
    if output_path is None:

        # Init error message
        instance_name = "software" if is_software else "file"
        error_str = f"\nERROR in find_file(): no valid path found for {instance_name} '{name}':"
        error_str += "\nPath to file not found among: "

        # List failed candidates
        for candidate_path in path_list:
            error_str += f"\n - '{candidate_path}'"
        if is_software:
            error_str += "\nCommand not found in the system PATH among: "
            for candidate_path in path_list:
                error_str += f"\n - '{os.path.basename(candidate_path)}'"

        # Add recommendaiton
        if is_software:
            error_str += f"\n   -> Please install software '{name}' and provide the path to its executable file or add it to system PATH."

        # Add description
        if description is not None:
            error_str += f"\nDescription: \n{description}"
        raise ValueError(error_str)

    # Return first found valid path
    return output_path
    
    



"""
NOTE about the 'setup.py' deprecation.
Despite the fact that 'setup.py' is now deprecated and replaced by 'pyproject.toml', it is still required to include C++ modules in pip packages.
Here is a minimal 'setup.py' that includes the C++ code in the package to complement the 'pyproject.toml' (this is fucked up)
"""

# Imports ----------------------------------------------------------------------
from setuptools import setup, Extension

# Extensions -------------------------------------------------------------------
# Define extension (C++ code that need to be compiled)
compute_weights_ext = Extension(
    'rsalor.weights.lib_computeWeightsBackend', # name
    sources=[ # .cpp files
        'rsalor/weights/computeWeightsBackend.cpp',
        'rsalor/weights/msa.cpp',
    ],
    include_dirs=[ # .h directories
        'rsalor/weights/include',
    ],
    extra_compile_args=['-std=c++11', '-O3'],  # optimization and other flags
    extra_link_args=['-O3'],
    language='c++',
)

# Setup ------------------------------------------------------------------------
setup(
    ext_modules = [compute_weights_ext],
)



# Imports ----------------------------------------------------------------------
import os.path
from abc import ABC, abstractmethod
from typing import Dict, Union


# Abstract RSASolver class -----------------------------------------------------
class RSASolver(ABC):
    """
    Abstract container class for RSASolver: to compute RSA (Relative Solvent Accessibility) from a PDB file.

    usage:
        rsa_map = RSASolver('./soft/software_executable').run('./my_pdb.pdb')
    """

    # Constants ----------------------------------------------------------------
    COMMENT_CHAR = "#"

    # Constructor --------------------------------------------------------------
    def __init__(
            self,
            executable_path: Union[None, str]=None,
            verbose: bool=False,
        ):
        self.executable_path = executable_path
        self.verbose = verbose
    
    # Methods ------------------------------------------------------------------
    def __str__(self) -> str:
        return "RSASolver['AbstractSolver']"
    
    def run(
            self,
            pdb_path: str,
            rsa_cache_path: Union[None, str]=None,
        ) -> Dict[str, float]:
        """Compute RSA by running the solver or using the caced file.

        args:
        pdb_path (str):         path to PDB file
        rsa_cache_path (str):   path to/from where save/read RSA (if file exists, solver execution will be skipped and output directly read from file)

        output:
        {resid: str => RSA: float}     (such as {'A13': 48.57, ...})
        """
        
        # Parse RSA if cache file exists
        if rsa_cache_path is not None and os.path.isfile(rsa_cache_path):
            if self.verbose:
                print(f" * read RSA values from rsa_cache_path '{rsa_cache_path}'")
            rsa_map = self.read(rsa_cache_path)
            return rsa_map
        
        # PDB file Guardians
        assert os.path.isfile(pdb_path), f"ERROR in {self}.execute_solver(): pdb_path='{pdb_path}' file does not exists."
        assert pdb_path.endswith(".pdb"), f"ERROR in {self}.execute_solver(): pdb_path='{pdb_path}' should be a '.pdb' file."

        # Run solver
        if self.verbose:
            print(f" * execute RSA solver: {self}")
        rsa_map = self.execute_solver(pdb_path)

        # Write RSA map
        if rsa_cache_path is not None and not os.path.isfile(rsa_cache_path):
            if self.verbose:
                print(f" * save RSA values in rsa_cache_path '{rsa_cache_path}'")
            self.write(rsa_cache_path, rsa_map)

        # Return
        return rsa_map

    @abstractmethod
    def execute_solver(self, pdb_path: str) -> Dict[str, float]:
        """Compute RSA by running the solver.

        args:
        pdb_path (str):         path to PDB file   

        output:
        {resid: str => RSA: float}     (such as {'A13': 48.57, ...})
        """
        pass

    def read(self, file_path: str) -> Dict[str, float]:
        """Read rsa_map file and return RSA mapping: {resid: str => RSA: float}."""

        # Guardians
        assert os.path.isfile(file_path), f"ERROR in {self}.read(): file file_path='{file_path}' does not exist."

        # Parse and return
        rsa_map: Dict[str, float] = {}
        with open(file_path, "r") as fs:
            lines = [line.split() for line in fs.readlines() if len(line) > 2 and line[0] != self.COMMENT_CHAR]
        for line in lines:
            resid, rsa = line[0], line[1]
            rsa_map[resid] = float(rsa)

        # Guardian and return
        assert len(rsa_map) > 0, f"ERROR in {self}.read(): No RSA data found in file_path='{file_path}'."
        return rsa_map
    
    def write(self, file_path: str, rsa_map: Dict[str, float]) -> "RSASolver":
        """Write rsa_map to a file."""

        # Guardians
        file_path = os.path.abspath(file_path)
        assert os.path.isdir(os.path.dirname(file_path)), f"ERROR in {self}.write(): directory of file_path='{file_path}' does not exists."

        # Stringity
        rsa_map_str = "\n".join(f"{resid} {rsa}" for resid, rsa in rsa_map.items())

        # Write
        with open(file_path, "w") as fs:
            fs.write(rsa_map_str)

        return self



# Imports ----------------------------------------------------------------------
import os.path
from typing import Dict
from Bio.PDB import PDBParser
from Bio.PDB.SASA import ShrakeRupley
from rsalor.sequence import AminoAcid
from rsalor.rsa.rsa_solver import RSASolver

# RSAMuSiC ---------------------------------------------------------------------
class RSABiopython(RSASolver):
    """
    RSABiopython(): Solver for RSA (Relative Solvent Accessibility) from a '.pdb' file using python package biopython.
    Uses the “rolling ball” algorithm developed by Shrake & Rupley algorithm
        doc: https://biopython.org/docs/dev/api/Bio.PDB.SASA.html
    
    usage:
        rsa_map = RSABiopython().run('./my_pdb.pdb')
    """

    # Constants ----------------------------------------------------------------
    # Taken from https://pmc.ncbi.nlm.nih.gov/articles/PMC3836772/#pone-0080635-t001
    MAX_SURFACE_MAP = {
        "ALA": 1.29,
        "ARG": 2.74,
        "ASN": 1.95,
        "ASP": 1.93,
        "CYS": 1.67,
        "GLN": 2.23,
        "GLU": 2.25,
        "GLY": 1.04,
        "HIS": 2.24,
        "ILE": 1.97,
        "LEU": 2.01,
        "LYS": 2.36,
        "MET": 2.24,
        "PHE": 2.40,
        "PRO": 1.59,
        "SER": 1.55,
        "THR": 1.55,
        "TRP": 2.85,
        "TYR": 2.63,
        "VAL": 1.74,
    }
    MAX_SURFACE_DEFAULT = 2.01 # mean value

    # Methods ------------------------------------------------------------------
    def __str__(self) -> str:
        return "RSASolver['biopython' (Shrake & Rupley algorithm)]"
    
    def execute_solver(self, pdb_path: str) -> Dict[str, float]:
        """Compute RSA by running biopython python package: Bio.PDB.SASA: ShrakeRupley
            doc: https://biopython.org/docs/dev/api/Bio.PDB.SASA.html

        args:
        pdb_path (str):         path to PDB file   

        output:
        {resid: str => RSA: float}     (such as {'A13': 48.57, ...})
        """

        # Parse PDB file
        pdb_name = os.path.basename(pdb_path).removesuffix(".pdb")
        pdb_parser = PDBParser(QUIET=True)
        structure = pdb_parser.get_structure(pdb_name, pdb_path)

        # Compute ASA
        shrake_rupley = ShrakeRupley(
            #probe_radius=1.40, # radius of the probe in A. Default is 1.40, roughly the radius of a water molecule.
            #n_points=200,      # resolution of the surface of each atom. Default is 100. A higher number of points results in more precise measurements, but slows down the calculation.
            #radii_dict=None,   # user-provided dictionary of atomic radii to use in the calculation. Values will replace/complement those in the default ATOMIC_RADII dictionary.
        )
        shrake_rupley.compute(structure, level="R")

        # Convert to RSA and format
        rsa_map: Dict[str, float] = {}
        for chain_obj in structure[0]:
            chain = chain_obj.id
            chain_structure = structure[0][chain]
            for residue in chain_structure:
                
                # Find 'resid' = {chain}{res_position}
                (res_insertion, res_id, res_alternate_location) = residue.id
                resid = f"{chain}{res_insertion}{res_id}".replace(" ", "")

                # Get AA 3-letter code and standardize if required
                aa_three = residue.resname
                aa_three = AminoAcid._NON_STANDARD_AAS.get(aa_three, aa_three)

                # Get RSA
                asa = residue.sasa
                if isinstance(asa, float):
                    rsa_map[resid] = asa / self.get_max_surf(aa_three)
        return rsa_map

    @classmethod
    def get_max_surf(cls, aa_three: str) -> float:
        return cls.MAX_SURFACE_MAP.get(aa_three, cls.MAX_SURFACE_DEFAULT)
    